/* gok-key.c
*
* Copyright 2001,2002 Sun Microsystems, Inc.,
* Copyright 2001,2002 University Of Toronto
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, write to the
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "gok-key.h"
#include "gok-keyboard.h"
#include "gok-log.h"
#include "gok-modifier.h"
#include "gok-scanner.h"
#include "gok-feedback.h"
#include "gok-data.h"
#include "gok-branchback-stack.h"

gchar m_TextEmpty[]="";

/**
* gok_key_new
* @pKeyPrevious: Pointer to the previous key in the list of keys.
* @pKeyNext: Pointer to the next key in the list
* @pKeyboard: Pointer to the keyboard.
*
* Allocates memory for a new key and initializes the GokKey structure.
* Returns a pointer to the new key, NULL if it can't be created.
* Call gok_key_delete on this when done with it.
*
* returns: A pointer to the new key, NULL if it wasn't created.
**/
GokKey* gok_key_new (GokKey* pKeyPrevious, GokKey* pKeyNext, GokKeyboard* pKeyboard)
{
	GokKey* pgok_key_new;

	g_assert (pKeyboard != NULL);

	/* allocate memory for the new key structure */
	pgok_key_new= (GokKey*) g_malloc(sizeof(GokKey));

	/* add the new key into the next/previous key list */
	if (pKeyboard->pKeyFirst == NULL)
	{
		pKeyboard->pKeyFirst = pgok_key_new;
	}

	if (pKeyPrevious != NULL)
	{
		pKeyPrevious->pKeyNext = pgok_key_new;
	}

	if (pKeyNext != NULL)
	{
		pKeyNext->pKeyPrevious = pgok_key_new;
	}

	/* initialize the data members of the structure */
	pgok_key_new->ImageCode = KEYDISPLAY_TEXT;
	pgok_key_new->Type = KEYTYPE_NORMAL;
	pgok_key_new->Style = KEYTYPE_NORMAL;
	pgok_key_new->pLabel = NULL;
	pgok_key_new->Target = NULL;
	pgok_key_new->pOutput = NULL;
	pgok_key_new->pOutputWrapperPre = NULL;
	pgok_key_new->pOutputWrapperPost = NULL;
	pgok_key_new->pButton = NULL;
	pgok_key_new->ModifierName = NULL;
	pgok_key_new->Top = 0;
	pgok_key_new->Bottom = 0;
	pgok_key_new->Left = 0;
	pgok_key_new->Right = 0;
	pgok_key_new->TopWin = 0;
	pgok_key_new->BottomWin = 0;
	pgok_key_new->LeftWin = 0;
	pgok_key_new->RightWin = 0;
	pgok_key_new->pMimicKey = NULL;
	pgok_key_new->pKeyNext = pKeyNext;
	pgok_key_new->pKeyPrevious = pKeyPrevious;
	pgok_key_new->pNodeAccessible = NULL;
	pgok_key_new->FontSizeGroup = FONT_SIZE_GROUP_UNDEFINED;
	pgok_key_new->FontSize = -1; /* not set yet */
	pgok_key_new->State = GTK_STATE_NORMAL;
	pgok_key_new->StateWhenNotFlashed = GTK_STATE_NORMAL;
	pgok_key_new->pGeneral = NULL;
	
	return pgok_key_new;
}

/**
* gok_key_delete
* @pKey: Pointer to the key that gets deleted.
* @pKeyboard: Pointer to the keyboard that contains the key (can be NULL). If
* pKeyboard is not NULL then the key is unhooked from the keyboard.
* @bDeleteButton: Flag that determines if the GTK button associated with
* the key should also be deleted. This should be set to TRUE if the key is
* deleted while the program is running. At the end of the program, when the
* GOK window is destroyed and the GTK buttons are destroyed, this should
* be set to FALSE.
* 
* Deletes the given key. This must be called on every key that has been created.
* Don't use the given key after calling this.
* This unhooks the key from the next/previous list of keys.
**/
void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButton)
{
	/* Don't call gok_spy_free on the NodeAccessible since this will be done by the keyboard. */
	GokKeyLabel* pLabel;
	GokKeyLabel* pLabelTemp;
	
	if (pKey == NULL)
	{
		return;
	}

	/* if this key is flashing then turn off the flashing */
	if (gok_feedback_get_key_flashing() == pKey)
	{
		gok_feedback_timer_stop_key_flash();
	}
	
	/* unhook the key from the keyboard */
	if ((pKeyboard != NULL) &&
		(pKeyboard->pKeyFirst == pKey))
	{
		pKeyboard->pKeyFirst = pKey->pKeyNext;
	}
	
	if (pKey->pKeyPrevious != NULL)
	{
		pKey->pKeyPrevious->pKeyNext = pKey->pKeyNext;
	}

	if (pKey->pKeyNext != NULL)
	{
		pKey->pKeyNext->pKeyPrevious = pKey->pKeyPrevious;
	}

	/* delete all the key's labels */
	pLabel = pKey->pLabel;
	while (pLabel != NULL)
	{
		pLabelTemp = pLabel;
		pLabel = pLabel->pLabelNext;
		gok_keylabel_delete (pLabelTemp);
	}

	/* delete the branch target string */
	if (pKey->Target != NULL)
	{
		g_free (pKey->Target);
	}

	/* delete the GTK button associated with the key */
	if ((pKey->pButton != NULL) &&
		(bDeleteButton == TRUE))
	{
		gtk_widget_destroy (pKey->pButton);
	}
	
	/* delete all the outputs associated with the key */
	gok_output_delete_all (pKey->pOutput);
	gok_output_delete_all (pKey->pOutputWrapperPre);
	gok_output_delete_all (pKey->pOutputWrapperPost);

	g_free (pKey);
}

/**
* gok_key_initialize
* @pKey: Pointer to the key that's getting initialized.
* @pNode: Pointer to the XML node that contains the key data.
*
* returns: TRUE if the key was initialized, FALSE if not.
**/ 
gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
{
	xmlChar* pStringAttributeValue;
	xmlNode* pNodeKeyChild;
	xmlNode* pNodeWrapperChild;
	GokOutput* pNewOutput;
	GokOutput* pOutputTemp;

	g_assert (pKey != NULL);
	g_assert (pNode != NULL);

	/* type of key (normal, branch etc.) */
	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "type");
	if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"normal") == 0)
	{
		pKey->Type = KEYTYPE_NORMAL;
		pKey->Style = KEYSTYLE_NORMAL;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *) "modifier") == 0)
	{
		pKey->Type = KEYTYPE_MODIFIER;
		pKey->Style = KEYSTYLE_NORMAL;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *) "branch") == 0)
	{
		pKey->Type = KEYTYPE_BRANCH;
		pKey->Style = KEYSTYLE_BRANCH;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchBack") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHBACK;
		pKey->Style = KEYSTYLE_BRANCHBACK;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchAlphabet") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHALPHABET;
		pKey->Style = KEYSTYLE_BRANCHALPHABET;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchToolbars") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHTOOLBARS;
		pKey->Style = KEYSTYLE_BRANCHTOOLBARS;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchMenus") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHMENUS;
		pKey->Style = KEYSTYLE_BRANCHMENUS;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchGUI") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHGUI;
		pKey->Style = KEYSTYLE_BRANCHGUI;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchEditText") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHEDITTEXT;
		pKey->Style = KEYSTYLE_BRANCHEDITTEXT;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"settings") == 0)
	{
		pKey->Type = KEYTYPE_SETTINGS;
		pKey->Style = KEYSTYLE_SETTINGS;
	}
	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchWindows") == 0)
	{
		pKey->Type = KEYTYPE_BRANCHWINDOWS;
		pKey->Style = KEYSTYLE_BRANCH;
	}

	/* branch target */
	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "target");
	{
		if (pStringAttributeValue != NULL)
		{
			pKey->Target = pStringAttributeValue;
		}
	}

	/* font size group*/
	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "fontsizegroup");
	if (pStringAttributeValue != NULL)
	{
		pKey->FontSizeGroup = atoi (pStringAttributeValue);
	}

	/* location of key (top, bottom, left & right) */
	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "top");
	if (pStringAttributeValue != NULL)
	{
		pKey->Top = atoi (pStringAttributeValue);
	}
	else
	{
		gok_log_x ("Can't find 'top' attribute in gok_key_initialize!\n");
		return FALSE;
	}

	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "bottom");
	if (pStringAttributeValue != NULL)
	{
		pKey->Bottom = atoi (pStringAttributeValue);
	}
	else
	{
		gok_log_x ("Warning: Can't find 'bottom' attribute in gok_key_initialize!\n");
		return FALSE;
	}

	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "left");
	if (pStringAttributeValue != NULL)
	{
		pKey->Left = atoi (pStringAttributeValue);
	}
	else
	{
		gok_log_x ("Warning: Can't find 'left' attribute in gok_key_initialize!\n");
		return FALSE;
	}

	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "right");
	if (pStringAttributeValue != NULL)
	{
		pKey->Right = atoi (pStringAttributeValue);
	}
	else
	{
		gok_log_x ("Warning: Can't find 'right' attribute in gok_key_initialize!\n");
		return FALSE;
	}

	/* is this a 'modifier' key? */
	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "modifier");
	if (pStringAttributeValue != NULL)
	{
		/* add the modifier to the list of modifiers */
		gok_modifier_add (pStringAttributeValue);
		
		/* store the name of the modifier on the key */
		pKey->ModifierName = (gchar*)g_malloc (strlen (pStringAttributeValue) + 1);
		strcpy (pKey->ModifierName, pStringAttributeValue);

		/* what type of modifier is this? */
		pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "modifiertype");
		if (pStringAttributeValue != NULL)
		{
			if (strcmp (pStringAttributeValue, "toggle") == 0)
			{
				gok_modifier_set_type (pKey->ModifierName, MODIFIER_TYPE_TOGGLE);
			}
		}
	}

	/* get child elements of the key */
	pNodeKeyChild = pNode->xmlChildrenNode;
	while (pNodeKeyChild != NULL)
	{
		/* key label */
		if (xmlStrcmp (pNodeKeyChild->name, (const xmlChar *)"label") == 0)
		{
			gok_keylabel_new (pKey, xmlNodeGetContent (pNodeKeyChild), xmlGetProp (pNodeKeyChild, (const xmlChar *) "modifier"));
		}

		/* output */
		else if (xmlStrcmp (pNodeKeyChild->name, (const xmlChar *)"output") == 0)
		{
			pNewOutput = gok_output_new_from_xml (pNodeKeyChild);
			if (pNewOutput != NULL)
			{
				if (pKey->pOutput == NULL)
				{
					pKey->pOutput = pNewOutput;
				}
				else
				{
					pOutputTemp = pKey->pOutput;
					while (pOutputTemp->pOutputNext != NULL)
					{
						pOutputTemp = pOutputTemp->pOutputNext;
					}
					pOutputTemp->pOutputNext = pNewOutput;
				}
			}
		}
		
		/* wrapper */
		else if (xmlStrcmp (pNodeKeyChild->name, (const xmlChar *)"wrapper") == 0)
		{
			/* is this key a modifier? */
			if (pKey->ModifierName == NULL)
			{
				gok_log_x ("Key '%s' is not modifier. Can't add wrapper!\n", gok_key_get_label (pKey));
			}
			else
			{
				/* get output of the wrapper */
				pNodeWrapperChild = pNodeKeyChild->xmlChildrenNode;
				while (pNodeWrapperChild != NULL)
				{
					if (xmlStrcmp (pNodeWrapperChild->name, (const xmlChar *)"output") == 0)
					{
						pNewOutput = gok_output_new_from_xml (pNodeWrapperChild);
						if (pNewOutput != NULL)
						{
							/* is this a 'pre' or 'post' wrapper? */
							pStringAttributeValue = xmlGetProp (pNodeKeyChild, (const xmlChar *) "type");
							if (pStringAttributeValue == NULL)
							{
								gok_log_x ("Wrapper does not have a type!\n");
							}
							else
							{
								if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"pre") == 0)
								{
									pKey->pOutputWrapperPre = pNewOutput;
								}
								else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"post") == 0)
								{
									pKey->pOutputWrapperPost = pNewOutput;
								}
								else
								{
									gok_log_x ("Wrapper type '%s' is invalid!\n", pStringAttributeValue);
								}
							}
						}
					}
					pNodeWrapperChild = pNodeWrapperChild->next;
				}
			}
		}
		
		pNodeKeyChild = pNodeKeyChild->next;
	}

	return TRUE;
}

/**
* gok_keylabel_new
* @pKey: Pointer to the key that gets the new label.
* @pLabelText: Text string for this label.
* @pModifierText: Modifier text string (e.g. "shift" or "ctrl").
*
* Allocates memory for a new key label and initializes the GokKeyLabel structure.
* Returns a pointer to the new key label, NULL if it can't be created.
* Add this label to a key so it will be deleted when the key is deleted.
*
* returns: A pointer to the new key label, NULL if it wasn't created.
**/
GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, gchar* pModifierText)
{
	GokKeyLabel* pNewLabel;
	GokKeyLabel* pKeyLabel;
	
	/* allocate memory for the new label structure */
	pNewLabel = (GokKeyLabel*) g_malloc(sizeof(GokKeyLabel));
	
	pNewLabel->Text = m_TextEmpty;
	pNewLabel->Modifier = m_TextEmpty;
	pNewLabel->pLabelNext = NULL;

	if (pLabelText != NULL)
	{
		pNewLabel->Text = (gchar*)g_malloc (strlen(pLabelText) + 1);
		strcpy (pNewLabel->Text, pLabelText);
	}
	
	if (pModifierText != NULL)
	{
		pNewLabel->Modifier = (gchar*)g_malloc (strlen(pModifierText) + 1);
		strcpy (pNewLabel->Modifier, pModifierText);
	}
	
	if (pKey != NULL)
	{
		pKeyLabel = pKey->pLabel;
		if (pKeyLabel == NULL)
		{
			pKey->pLabel = pNewLabel;
		}
		else
		{
			while (pKeyLabel->pLabelNext != NULL)
			{
				pKeyLabel = pKeyLabel->pLabelNext;
			}
			pKeyLabel->pLabelNext = pNewLabel;
		}
	}
	
	return pNewLabel;
}

/**
* gok_keylabel_delete
* @pKeyLabel: Pointer to the key label that will be deleted.
**/ 
void gok_keylabel_delete (GokKeyLabel* pKeyLabel)
{
	if (pKeyLabel != NULL)
	{
		if (pKeyLabel->Text != m_TextEmpty)
		{
			g_free (pKeyLabel->Text);
		}
		if (pKeyLabel->Modifier != m_TextEmpty)
		{
			g_free (pKeyLabel->Modifier);
		}
		g_free (pKeyLabel);
	}
}

/**
* gok_key_add_label
* @pKey: Pointer to the key that's gets the new label.
* @pLabel: Pointer to the label text.
* @pModifier: Pointer to the 'modifier' for the key's label.
*
* Adds a label to the key. This allocates memory for the label that will
* be freed in gok_key_delete.
*
* returns: TRUE if the key was initialized, FALSE if not.
**/ 
gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, gchar* pModifier)
{
	GokKeyLabel* pKeyLabel;
		
	g_assert (pKey != NULL);
	g_assert (pLabelText != NULL);
	
	if (pKey->pLabel == NULL)
	{
		gok_keylabel_new (pKey, pLabelText, pModifier);
	}
	else
	{
		/* if the key already has a label with the same modifier then change it */
		pKeyLabel = pKey->pLabel;
		while (pKeyLabel != NULL)
		{
			if (strcmp (pKeyLabel->Modifier, pModifier) == 0)
			{
				g_free (pKeyLabel->Text);
				pKeyLabel->Text = (gchar*)g_malloc (strlen(pLabelText) + 1);
				strcpy (pKeyLabel->Text, pLabelText);
				
				break;
			}
			pKeyLabel = pKeyLabel->pLabelNext;
		}
		
		if (pKeyLabel == NULL)
		{
			gok_keylabel_new (pKey, pLabelText, pModifier);
		}
	}
		
	return TRUE;
}

/**
* gok_key_change_label
* @pKey: Pointer to the key that gets the new label.
* @LabelText: The new label text.
*
* Changes the label displayed on the gok key.
**/
void gok_key_change_label (GokKey* pKey, gchar* LabelText)
{
	g_assert (pKey != NULL);
	g_assert (LabelText != NULL);
	
	gok_key_add_label (pKey, LabelText, MODIFIER_NORMAL);
	
	/* change the label text */
	gok_key_set_button_label (pKey, LabelText);

	/* set the label 'name' */
	gok_key_set_button_name (pKey);	
}

/**
* gok_key_update_label
* @pKey: Pointer to the key that gets an updated label.
*
* Changes the key's label if the modifier state has changed.
**/
void gok_key_update_label (GokKey* pKey)
{
	gchar* pNewLabelText;
	const gchar* pOldLabelText;
	GtkWidget* pButtonLabel;
	
	if (pKey->pButton == NULL)
	{
		return;
	}

	/* get the key's label text */
	pNewLabelText = gok_key_get_label (pKey);
	
	/* get the current text displayed on the key */
	pButtonLabel = ((GokButton*)pKey->pButton)->pLabel;
	pOldLabelText = gtk_label_get_text (GTK_LABEL(pButtonLabel));
	if (pOldLabelText == NULL)
	{
		return;
	}
	
	/* if the new text is different from the old text then change it */
	if (strcmp (pNewLabelText, pOldLabelText) != 0)
	{
		gtk_label_set_text (GTK_LABEL(pButtonLabel), pNewLabelText);
	}
}

/**
* gok_key_get_label
* @pKey: Pointer to the key that you want the label for.
*
* Returns: A pointer to the label's text string, NULL if no label.
**/
gchar* gok_key_get_label (GokKey* pKey)
{
	gchar* pTextReturned;
	gchar* pToken;
	gchar buffer[150];
	GokKeyLabel* pKeyLabel;
	gboolean bFoundLabel;
	
	g_assert (pKey != NULL);
	
	pTextReturned = m_TextEmpty;
	
	/* get the 'normal' label */
	pKeyLabel = pKey->pLabel;
	while (pKeyLabel != NULL)
	{
		if (strcmp (pKeyLabel->Modifier, MODIFIER_NORMAL) == 0)
		{
			pTextReturned = pKeyLabel->Text;
			break;
		}
		pKeyLabel = pKeyLabel->pLabelNext;
	}

	/* are any modifiers turned on? */
	if (gok_modifier_get_normal() == TRUE)
	{
		/* no modifiers, so just return the 'normal' label */
		return pTextReturned;
	}
	
	/* find the correct label text for the current modifiers */
	/* look through all the labels for one that best matches the modifiers */
	bFoundLabel = FALSE;
	pKeyLabel = pKey->pLabel;
	while (pKeyLabel != NULL)
	{
		/* does this label apply to more than one modifier? (e.g. ctrl+shift) */
		if (strchr (pKeyLabel->Modifier, '+') != 0)
		{
			/* check if all these modifiers are on */
			strcpy (buffer, pKeyLabel->Modifier);
			pToken = strtok (buffer, "+");
			while (pToken != NULL)
			{
				if (gok_modifier_get_state (pToken) == MODIFIER_STATE_OFF)
				{
					break;
				}
				
				pToken = strtok (NULL, "+");
				
				if (pToken == NULL)
				{
					/* all of the modifiers are on, return this label text */
					return pKeyLabel->Text;
				}
			}
		}
		else if ((bFoundLabel == FALSE) &&
				(strcmp (pKeyLabel->Modifier, MODIFIER_NORMAL) != 0))
		{
			/* the first modifier state that matches will be used */
			if (gok_modifier_get_state (pKeyLabel->Modifier) != MODIFIER_STATE_OFF)
			{
				pTextReturned = pKeyLabel->Text;
				bFoundLabel = TRUE;
			}
		}
		pKeyLabel = pKeyLabel->pLabelNext;
	}
	
	return pTextReturned;
}

/**
* gok_key_set_output
* @pKey: Pointer to the key that's gets the new output.
* @Type: Type of output (e.g. keysym or keycode)
* @pName: Pointer to the name string.
* @Flag: Type of key synth output (if relevant)
*
* Sets the output for the key. This allocates memory for the output that will
* be freed in gok_key_delete.
**/ 
void gok_key_set_output (GokKey* pKey, gint Type, gchar* pName, AccessibleKeySynthType Flag)
{
	g_assert (pKey != NULL);
	
	gok_output_delete_all (pKey->pOutput);
	if (pName != NULL)
	{
		pKey->pOutput = gok_output_new (Type, pName, Flag);
	}
}

/**
* gok_key_add_output
* @pKey: Pointer to the key that's gets the new output.
* @Type: Type of output (e.g. keysym or keycode)
* @pName: Pointer to the name string.
* @Flag: Type of key synth output (if relevant)
*
* Adds output for the key. This allocates memory for the output that will
* be freed in gok_key_delete.
**/ 
void gok_key_add_output (GokKey* pKey, gint Type, gchar* pName, AccessibleKeySynthType Flag)
{
	GokOutput* pOutputPrevious;
	GokOutput* pOutputTemp = NULL;

	gok_log_enter();
	
	g_assert (pKey != NULL);
	
	if (pName != NULL)
	{
		pOutputPrevious = pOutputTemp;
		pOutputTemp = pKey->pOutput;
		while (pOutputTemp != NULL)
		{
			pOutputPrevious = pOutputTemp;
			pOutputTemp = pOutputTemp->pOutputNext;
		}
		
		pOutputTemp = gok_output_new (Type, pName, Flag);
		
		if (pOutputPrevious == NULL)
		{ 	
	    	pKey->pOutput = pOutputTemp; 
	    }
	    else
	    {
	    	pOutputPrevious->pOutputNext = pOutputTemp;
	    }
	}

	gok_log_leave();
}

/**
* gok_key_set_button_name
* @pKey: Pointer to the key that gets the new 'name'.
*
* Sets the 'name' of the key's label. The 'name' is used to determine
* the key/label colors from the .rc file. This must be called for every key
* after it's created and after the label name has been changed.
**/
void gok_key_set_button_name (GokKey* pKey)
{
	gchar styleButton[200];
	gchar styleText[200];

	switch (pKey->Style)
	{
		case KEYSTYLE_NORMAL:
			strcpy (styleButton, "StyleButtonNormal");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCH:
			strcpy (styleButton, "StyleButtonBranch");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHBACK:
			/* disable the button if we can't branch back */
			if (gok_branchbackstack_is_empty() == TRUE)
			{
				strcpy (styleButton, "StyleButtonDisabled");
				strcpy (styleText, "StyleTextDisabled");
			}
			else
			{
				strcpy (styleButton, "StyleButtonBranchBack");
				strcpy (styleText, "StyleTextNormal");
			}
			break;

		case KEYSTYLE_GENERALDYNAMIC:
			strcpy (styleButton, "StyleButtonGeneralDynamic");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHMENUS:
			strcpy (styleButton, "StyleButtonBranchMenus");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHMENUITEMS:
			strcpy (styleButton, "StyleButtonBranchMenuItems");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_MENUITEM:
			strcpy (styleButton, "StyleButtonMenuItem");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHTOOLBARS:
			strcpy (styleButton, "StyleButtonBranchToolbars");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHGUI:
			strcpy (styleButton, "StyleButtonBranchGUI");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHGUIACTIONS:
		case KEYSTYLE_EDITABLETEXTACTION:
			strcpy (styleButton, "StyleButtonBranchGuiActions");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_BRANCHALPHABET:
			strcpy (styleButton, "StyleButtonBranchAlphabet");
			strcpy (styleText, "StyleTextNormal");
			break;

		case KEYSTYLE_SETTINGS:
			strcpy (styleButton, "StyleButtonSettings");
			strcpy (styleText, "StyleTextNormal");
			break;
					
		case KEYSTYLE_WORDCOMPLETE:
			strcpy (styleButton, "StyleButtonWordComplete");
			strcpy (styleText, "StyleTextWordComplete");
			break;
		
		default:
			gok_log_x ("Warning: default hit in gok_key_set_button_name! Key = %s\n", gok_key_get_label (pKey));
			break;
	}
	
	/* set the 'name' of the button */
	gtk_widget_set_name (pKey->pButton, styleButton); 

	/* set 'name' of the button text */
	g_assert (((GokButton*)pKey->pButton)->pLabel != NULL);
	gtk_widget_set_name (((GokButton*)pKey->pButton)->pLabel, styleText);
}

/**
* gok_key_calculate_font_size
* @pKey: Pointer to the key that we're getting the font size for.
* @bWidth: If TRUE then calculate the font size needed for the width
* of the key's label.
* @bHeight: If TRUE then calculate the font size needed for the height
* of the key's label.
* Note: Both bWidth and bHeight can be TRUE.
*
* Calculates the font size needed for the key's label to fill the key.
*
* Returns: The font size, in 1000s of a point (e.g. 9 point is 9000).
**/
int gok_key_calculate_font_size (GokKey* pKey, gboolean bWidth, gboolean bHeight)
{
	gint sizeFontWidth;
	gint sizeFontHeight;
	gint sizeFontReturned;
	gint sizeTemp;
	gint widthCell;
	gint widthTrial;
	gint heightCell;
	gint heightTrial;
	gint increment;
	
	/* at least one must be TRUE */
	g_assert ((bWidth == TRUE) || (bHeight == TRUE));

	/* store the current size of the key */
	sizeTemp = pKey->FontSize;

	/* start at a small font size and increase to a very large size */
	/* font sizes are expressed in 1000th of a point (1000 = 1 point) */
	/* start at 5 points and go up to 50 points in increments of 1 point */
	/* first check cell width */
	if (bWidth == TRUE)
	{
		/* get the size of a key cell*/
		widthCell = gok_data_get_key_width() - 5;
	
		increment = 1000;
		for (sizeFontWidth = 5000; sizeFontWidth < 200000; sizeFontWidth += increment)
		{
			/* set the key label to the trial size */
			gok_key_set_font_size (pKey, sizeFontWidth);
			
			/* get the size of the label at the trial size */
			widthTrial = gok_key_get_label_lengthpercell (pKey);
			
			/* stop if the label fits the key cell exactly */
			if (widthTrial == widthCell)
			{
				break;
			}
			
			/* is it bigger than the cell size? */
			else if (widthTrial > widthCell)
			{
				/* label is larger than a cell so decrease it a bit */
				sizeFontWidth -= increment;
				
				/* stop looking */
				break;
			}
			
			/* at very large font sizes, increase the increment we increase */
			if (sizeFontHeight > 50)
			{
				increment = 2000;
			}
			else if (sizeFontHeight > 100)
			{
				increment = 5000;
			}
		}
	}
	
	sizeFontReturned = sizeFontWidth;
	
	/* now check the cell height */
	if (bHeight == TRUE)
	{
		/* get the size of a key cell*/
		heightCell = gok_data_get_key_height() - 10;
	
		for (sizeFontHeight = 5000; sizeFontHeight < 200000; sizeFontHeight += increment)
		{
			/* set the key label to the trial size */
			gok_key_set_font_size (pKey, sizeFontHeight);
			
			/* get the size of the label at the trial size */
			heightTrial = gok_key_get_label_heightpercell (pKey);
			
			/* stop if the label fits the key cell exactly */
			if (heightTrial == heightCell)
			{
				break;
			}
			
			/* is it bigger than the cell size? */
			else if (heightTrial > heightCell)
			{
				/* label is larger than a cell so decrease it a bit */
				sizeFontHeight -= increment;
				
				/* stop looking */
				break;
			}
			
			/* at very large font sizes, increase the increment we increase */
			if (sizeFontHeight > 50)
			{
				increment = 2000;
			}
			else if (sizeFontHeight > 100)
			{
				increment = 5000;
			}
		}

		/* did we also get the font width? */
		if (bWidth == TRUE)
		{
			sizeFontReturned = (sizeFontHeight < sizeFontWidth) ? sizeFontHeight : sizeFontWidth;
		}
		else
		{
			sizeFontReturned = sizeFontHeight;
		}
	}
	
	/* restore key size to original size */
	if (sizeTemp != -1)
	{
		gok_key_set_font_size (pKey, sizeTemp);
	}
	else
	{
		pKey->FontSize = -1;
	}
	
	return sizeFontReturned;
}

/**
* gok_key_set_font_size
* @pKey: Pointer to the key that gets the new font size.
* @Size: Font size you want the key's text to be.
*
* Sets the font size for the key.
**/
void gok_key_set_font_size (GokKey* pKey, gint Size)
{
	PangoAttrList* pPangoAttrs;
	GokButton* pButton;
	gchar* pLabel;
	gchar SafeLabel[351]; /* will hold the HTML safe version of the label */
	gchar string [400];
	g_assert (pKey != NULL);
	g_assert (pKey->pButton != NULL);

	if (Size <= 0)
	{
		Size = 200;
	}
	
	if (pKey->pLabel == NULL)
	{
		return;
	}

	/* store the size on the key */
	pKey->FontSize = Size;
	
	if (strlen (gok_key_get_label (pKey)) > 300)
	{
		gok_log_x ("Warning: gok_key_set_font_size failed because key label too long.\n"); 
		return;
	}

	/* get an HTML save version of the label */
	pLabel = gok_key_get_label (pKey);
	gok_key_make_html_safe (pLabel, SafeLabel, 350);
	
	/* create a list of Pango attributes */
	sprintf (string, "<span size=\"%d\">%s</span>", Size, SafeLabel);
	if (pango_parse_markup (string, -1, 0, &pPangoAttrs, NULL, NULL, NULL) == TRUE)
	{
		/* add the Pango atrribute list to the key's label */
		pButton = (GokButton*)pKey->pButton;
		gtk_label_set_attributes (GTK_LABEL(pButton->pLabel), pPangoAttrs);
	}
	else
	{
		gok_log_x ("Warning: pango_parse_markup failed in gok_key_set_font_size! String = %s\n", string);
	}
}

/**
* gok_key_make_html_safe
* @pString: String that needs to be made HTML safe.
* @pSafeString: Pointer to a buffer that will hold the HTML safe string.
* @SafeStringLength: Length of the buffer that holds the HTML safe string.
*
* Converts a given string to an HTML safe string. This converts characters
* like '<' to &lt;
*
* Returns: TRUE if pSafeString is OK, FALSE if not.
**/
gboolean gok_key_make_html_safe (gchar* pString, gchar* pSafeString, gint SafeStringLength)
{
	gint i1, i2;
	
	for (i1 = 0, i2 = 0; i1 < strlen (pString); i1++)
	{
		if (pString[i1] == '<')
		{
			if (i2 > (SafeStringLength - 5))
			{
				gok_log_x ("pSafeString is too short to hold pString!");
				pSafeString[i2] = 0;
				return FALSE;
			}

			pSafeString[i2++] = '&';
			pSafeString[i2++] = 'l';
			pSafeString[i2++] = 't';
			pSafeString[i2++] = ';';
		}
		else if (pString[i1] == '>')
		{
			if (i2 > (SafeStringLength - 5))
			{
				gok_log_x ("pSafeString is too short to hold pString!");
				pSafeString[i2] = 0;
				return FALSE;
			}
			pSafeString[i2++] = '&';
			pSafeString[i2++] = 'g';
			pSafeString[i2++] = 't';
			pSafeString[i2++] = ';';
		}
		else if (pString[i1] == '&')
		{
			if (i2 > (SafeStringLength - 6))
			{
				gok_log_x ("pSafeString is too short to hold pString!");
				pSafeString[i2] = 0;
				return FALSE;
			}
			pSafeString[i2++] = '&';
			pSafeString[i2++] = 'a';
			pSafeString[i2++] = 'm';
			pSafeString[i2++] = 'p';
			pSafeString[i2++] = ';';
		}
		else if (pString[i1] == 34) /* double quote mark */
		{
			if (i2 > (SafeStringLength - 7))
			{
				gok_log_x ("pSafeString is too short to hold pString!");
				pSafeString[i2] = 0;
				return FALSE;
			}
			pSafeString[i2++] = '&';
			pSafeString[i2++] = 'q';
			pSafeString[i2++] = 'u';
			pSafeString[i2++] = 'o';
			pSafeString[i2++] = 't';
			pSafeString[i2++] = ';';
		}
		else
		{
			if (i2 > (SafeStringLength - 2))
			{
				gok_log_x ("pSafeString is too short to hold pString!");
				pSafeString[i2] = 0;
				return FALSE;
			}

			pSafeString[i2++] = pString[i1];
		}
	}
	
	pSafeString[i2] = 0;
	return TRUE;
}

/**
* gok_key_get_label_lengthpercell
* @pKey: Pointer to the key you want the to find the label length per cell.
*
* Calculates the length of the key's label. Some keys span more
* than one cell so divide the label length into the number of cells.
*
* Returns: The length of the key's label per cell.
**/
int gok_key_get_label_lengthpercell (GokKey* pKey)
{
	PangoLayout* pPangoLayout;
	PangoRectangle rectInk;
	PangoRectangle rectLogical;
	gint numCells;
	GtkLabel* pLabel;

	if ((pKey->pButton == NULL) ||
		(strlen (gok_key_get_label (pKey)) == 0))
	{
		return 0;
	}
	
	pLabel = GTK_LABEL(((GokButton*)pKey->pButton)->pLabel);
	g_assert (pLabel != NULL);
	
	pPangoLayout = gtk_label_get_layout (pLabel);
	pango_layout_get_pixel_extents (pPangoLayout, &rectInk, &rectLogical);
	
	numCells = pKey->Right - pKey->Left;
	g_assert (numCells > 0);
	return rectInk.width / numCells;
}

/**
* gok_key_get_label_heightpercell
* @pKey: Pointer to the key you want to find the hight per cell.
*
* Calculates the height of the key's label. Some keys span more
* than one cell so divide the label height into the number of cells.
*
* Returns: The height of the key's label per cell.
**/
int gok_key_get_label_heightpercell (GokKey* pKey)
{
	PangoLayout* pPangoLayout;
	PangoRectangle rectInk;
	PangoRectangle rectLogical;
	gint numCells;
	GtkLabel* pLabel;

	if ((pKey->pButton == NULL) ||
		(strlen (gok_key_get_label (pKey)) == 0))
	{
		return 0;
	}
	
	pLabel = GTK_LABEL(((GokButton*)pKey->pButton)->pLabel);
	g_assert (pLabel != NULL);
	
	pPangoLayout = gtk_label_get_layout (pLabel);
	pango_layout_get_pixel_extents (pPangoLayout, &rectInk, &rectLogical);
	
	numCells = pKey->Bottom - pKey->Top;
	g_assert (numCells > 0);
	return rectInk.height / numCells;
}

/**
* gok_key_set_button_label
* @pKey: Pointer to the key that will have it's button label changed.
* @LabelText: Text for the button.
*
* Changes the button label displayed on the key.
**/
void gok_key_set_button_label (GokKey* pKey, gchar* LabelText)
{
	GokButton* pButton;
	GtkWidget* pButtonLabel;

	g_assert (pKey != NULL);
	
	pButton = (GokButton*)pKey->pButton;
	if (pButton == NULL)
	{
		return;
	}

	pButtonLabel = pButton->pLabel;
	if (pButtonLabel == NULL)
	{
		return;
	}
	
	gtk_label_set_text (GTK_LABEL(pButtonLabel), LabelText);
}


/**
* gok_key_set_cells
* @pKey: Key that gets it's cells changed.
* @top: Top cell for the key.
* @bottom: Bottom cell for the key.
* @left: Left cell for the key.
* @right: Right cell for the key.
*
* Changes the cell coordinates (used by the editor).
**/
void gok_key_set_cells(GokKey* pKey, gint top, gint bottom, gint left, gint right)
{
	g_assert (pKey != NULL);
	g_assert (left < right);
	g_assert (top < bottom);
	pKey->Top = top;
	pKey->Bottom = bottom;
	pKey->Left = left;
	pKey->Right = right;
	
	/* update the button */
}

/**
* gok_key_duplicate
* @pKey: Pointer to the key that gets duplicated
*
* Not implemented yet.
*
* Returns: A pointer to the duplicate key, NULL if it was not created.
**/
GokKey* gok_key_duplicate(GokKey* pKey)
{
	return NULL;
}

